'''*****************************************
* turtlePython
* Name: Liz Courts
* E-mail: lizcourts97@unm.edu
* Course: CS491, Computational Fabrication
* Date: 2023-09-14
*
* A module to collect functions and classes for
* use in Processing for Python.
*
* REQUIREMENTS:
* - [Processing for Python](https://py.processing.org/), v. 3.5.4
* - [Turtle from Leah Buechley](https://leahbuechley.com/Turtle/index.html)
*
* INSTALLATION:
* # In your Processing sketch folder, there should be a 'libraries'
*   folder where other Processing libraries are stored. Create a
*   folder named 'turtlePython' in this directory.
* # Add this file to the 'libraries/turtlePython' directory.
* # At the beginning of your Processing sketch, add the following
*   lines:
*    sys.path.append(os.path.abspath('../libraries/turtlePython'))
*    import turtlePython as tp
* # Processing won't recognize the turtlePython module unless it's saved first.
*   Save the sketch, then run it.
*
* INSTRUCTIONS:
* - 't' in the functions indicate the Turtle object passed from
* the Processing program.
* - Production rules (LSystem.rules) are defined as dictionaries,
*   with values stored as a tuple. If the value is a list,
*   LSystem.iteration() will assume that it needs to pick a random
*   entry.
*
* EXAMPLE:
* lsys = tp.LSystem()
* lsys.rules['F'] = ['F-F+F-F']
* lsys.iteration()
* lsys.draw(t, lsys.computed_word, 10, 60, 1.0)
***********************************************'''

# we need our good friend random
import random

class LSystem(object):

	def __init__(self, axiom = 'F'):
		self.axiom = axiom
		self.rules = {
			'F': ['F'], # forward
			'X': ['X'], # do nothing
			'+': ['+'], # turn right
			'-': ['-'], # turn left
			'[': ['['], # push
			']': [']'], # pop
			'|': ['|'], # rotate 180
			'<': ['<'],
			'>': ['>'],
			'{': ['{'],
			'}': ['}'],
# TODO: define some default polygon shapes
			### polygons
			'o': ['o'],
			'u': ['u']
		}
		self.computed_word = self.axiom
		self.angle = 90
		self.length = 10
	
	# number of iterations to cycle through, output iterations to console
	# TODO: add option to find a random entry and keep it throughout the iteration(s)
	#       on a per-rule basis
	def iteration(self, iterations = -1, output = False):
		loop_count = 0
		
		while loop_count < iterations:
			new_word = ''
			c = ''
			i = 0
			
			while i < len(self.computed_word):
				c = self.computed_word[i]
				# find key in rules, pull random entry
				s = random.choice(self.rules.get(c, ''))
				new_word += s
				i += 1
		
			self.computed_word = new_word
			
			if output:
				print(str(loop_count+1) + ': ', self.computed_word)
			
			loop_count += 1
	
	# Turtle object, string, length, angle, scale_factor
	def draw(self, t, s = '', l = 0, a = '', scale_factor = 1.0, jitter = False):
		i = 0
		original_angle = a
		original_length = l
		
		stroke_width = l
		
		new_shape = createShape()
		
		if l==0:
			l = self.length * scale_factor
		else:
			l = l * scale_factor
			
		if not a:
			a = self.angle
			
		while i < len(s):
		
			if jitter:
				a = random.uniform(original_angle * 0.75, original_angle * 1.25)
				l = random.uniform(original_length * 0.75, original_length * 1.25)
		
			if s[i]=='+':
				t.right(a)
			elif s[i]=='-':
				t.left(a)
			elif s[i]=='|':
				t.right(180)
			elif s[i]=='[':
				t.push()
			elif s[i]==']':
				t.pop()
			elif s[i]=='>':
				#l = l * scale_factor
				l = l + 1
			elif s[i]=='<':
				#l = l / scale_factor
				l = l - 1
			elif s[i]=='#':
				stroke_width += 1
				strokeWeight(stroke_width)
			elif s[i]=='!':
				stroke_width -= 1
				if stroke_width < 1:
					stroke_width = 1
				strokeWeight(stroke_width)
			elif s[i]=='{':
				new_shape.beginShape()
			elif s[i]=='}':
				new_shape.endShape(CLOSE)
			elif s[i]=='F':
				t.forward(l)
			elif (s[i]=='f' or s[i]=='G'):
				t.penUp()
				t.forward(l)
				t.penDown()
			elif s[i]=='H':
				t.forward(l)
			elif s[i]=='u':
				leaf_size = self.shape('leaf')
				print(leaf_size)
				n = 0
				while n < leaf_size:
					t.forward(l)
					n += 1
			else:
				# TODO: implement additional interpretations
				#   from paulbourke.net/fractals/lsys
				# @ draw a dot with line width radius
				# { open polygon
				# } close polygon and fill
				# & swap + and -
				# ( decrement turning angle by turning angle increment
				# ) increment turning angle by turning angle increment
				pass
			i += 1
			
	def shape(self, name = 'leaf', startX = 0, startY = 0, l = 10):
		self.name = name
		self.startX = startX
		self.startY = startY
		self.l = l
		
		if name=='leaf':
			return 6
		else:
			pass
		
# pre-defined L-system classes with their production rules
# TODO: fix passing of angle? not sure if this is working right
class LSysKoch(LSystem):
	def __init__(self, axiom = 'F'):
		super(LSysKoch, self).__init__(axiom)
		self.rules['F'] = ['F-F++F-F']
		self.angle = 60
		
class LSysKochQIsland(LSystem):
	def __init__(self, axiom = 'F-F-F-F'):
		super(LSysKochQIsland, self).__init__(axiom)
		self.rules['F'] = ['F-F+F+FF-F-F+F']
		self.angle = 90
		
class LSysCCurve(LSystem):
	def __init__(self, axiom = 'F'):
		super(LSysCCurve, self).__init__(axiom)
		self.rules['F'] = ['+F--F+']
		self.angle = 45
		
class LSysDragonCurve(LSystem):
	def __init__(self, axiom = 'F'):
		super(LSysDragon, self).__init__(axiom)
		self.rules['F'] = ['F+H+']
		self.rules['H'] = ['-F-H']
		self.angle = 90
		
class LSysHilbertCurve(LSystem):
	def __init__(self, axiom = 'A'):
		super(LSysHilbertCurve, self).__init__(axiom)
		self.rules['A'] = ['+BF-AFA-FB+']
		self.rules['B'] = ['-AF+BFB+FA-']
		self.angle = 90
		
class LSysSierpinskiTriangle(LSystem):
	def __init__(self, axiom = 'F-F-F'):
		super(LSysSierpinskiTriangle, self).__init__(axiom)
		self.rules['F'] = ['F-H+F+H-F']
		self.rules['H'] = ['HH']
		self.angle = 120
		
class LSysSierpinskiGasket(LSystem):
	def __init__(self, axiom = 'H'):
		super(LSysSierpinskiGasket, self).__init__(axiom)
		self.rules['F'] = ['H-F-H']
		self.rules['H'] = ['F+H+F']
		self.angle = 60

class LSysPlant(LSystem):
	def __init__(self, axiom = 'X'):
		super(LSysPlant, self).__init__(axiom)
		self.rules['X'] = ['F+[[X]-X]-F[-FX]+X']
		self.rules['F'] = ['FF']
		self.angle = 25
		
		self.examples = {
			# "The Algorithmic Beauty of Plants", section 1.6
			# iterations, angle, axiom, rule
			'a': [5, 25.7, 'F', {'F': ['F[+F]F[-F]F']}],
			'aa': [5, 25.7, 'F', {'F': ['F[+F]F[-F]F', 'F[+F]F', 'F[-F]F']}],
			'b': [5, 20, 'F', {'F': ['F[+F]F[-F][F]']}],
			'c': [4, 22.5, 'F', {'F': ['FF-[-F+F+F]+[+F-F-F]']}],
			'd': [7, 20, 'X', {'X': ['F[+X]F[-X]+X'], 'F': ['FF']}],
			'e': [7, 25.7, 'X', {'X': ['F[+X][-X]FX'], 'F': ['FF']}],
			'f': [5, 22.5, 'X', {'X': ['F-[[X]+X]+F[+FX]-X'], 'F': ['FF']}],
		}

# Shapes

def solid_shape(l = 1, s = 4, stroke_color = [0,0,0], fill_color = [255,255,255]):
	pass

# Designs

def thing(t, l = 100, a = 90):
	t.forward(l)
	t.right(a)
	t.forward(l)
	t.right(a)
	t.forward(l / 2)
	t.right(a)
	t.forward(l / 2)
	t.right(a)
	t.forward(l)
	t.right(a)
	t.forward(l / 4)
	t.right(a)
	t.forward(l / 4)
	t.right(a)
	t.forward(l / 2)
		
def random_move(t, l = 20, a = 45):
	t.forward(random.randint(1,l))
	t.right(random.randint(-a, a))

# useful functions
def is_multiple(x, y):
	print(x, y % x)
	return x and (y % x) == 0

def shape(t, l = 100, sides = 4):

	x = 1

	while x <= sides:
		t.forward(l)
		t.right(360 / sides)
		x += 1

def poly(t, l = 1, a = 1):

	# figure out number of times to repeat
	i = 1
	rotate = False
	x = 0
	
	while not rotate:
		
		t.forward(l)
		t.right(a)
		
		x = (i * a) % 360
		
		if x == 0:
			rotate = True
		else:
			i += 1
			
		
def newpoly(t, l = 1, a = 1):

	i = 1
	rotate = False
	x = 0
	
	while not rotate:
		
		t.forward(l)
		t.right(a)
		t.forward(l)
		t.right(a * 2)
		
		x = (i * a) % 360
		
		if x == 0:
			rotate = True
		else:
			i += 1
			
def polyspi(t, l = 1, a = 1, i = 1, max = 500):

	if l >= max:
		pass
	else:
		t.forward(l)
		t.right(a)
		polyspi(t, l + i, a, i, max)
		
def inspi(t, l = 1, a = 1, i = 1):

	t.forward(l)
	t.right(a)
	
def poly_step(t, l = 1, a = 90):
	
	t.forward(l)
	t.right(a)
	
def poly_halt(t, l = 1, a = 90):

	r = 0
	t.forward(l)
	t.right(a)
	r += a

	while r % 360 > 0:
		t.forward(l)
		t.right(a)
		r += a
		
# recursive tree-drawing procedure
def branch(t, i = 0, l = 10, a = 45):
	if i==0:
		pass
	else:
		t.forward(l)
		t.left(a)
		branch(t, i-1, l * 1.5, a)
		t.right(a)
		t.right(a)
		branch(t, i-1, l * 1.5, a)
		t.left(a)
		t.back(l)